<div class="cockpit-logline" role="row">
  {{! placeholders for correct message alignment }}
  <div class="cockpit-log-warning" role="cell"></div>
  <span class="cockpit-log-message cockpit-logmsg-reboot" role="cell">{{message}}</span>
</div>
